import errno
import os
import Queue
import select
import sys

from .proc import Proc


class Worker(Proc):

    signal_map = {
        "tstp": "toggle_active",
        "term": "wind_down_gracefully",
        "quit": "wind_down_immediately"
    }
    source_handlers = {}
    outputs = {}

    def __init__(self, master):
        super(Worker, self).__init__()

        self.config = master.config
        self.redis = master.redis

        possible_channels = {
            "ready": master.ready_queue,
            "results": master.results_queue,
            "connection": master.conn
        }
        self.sources = {
            channel_name: possible_channels[channel_name]
            for channel_name in self.source_handlers.keys()
            if channel_name in possible_channels
        }
        self.handlers = {
            input_name: getattr(self, handler_name)
            for input_name, handler_name in self.source_handlers.iteritems()
            if getattr(self, handler_name, None)
        }
        self.outputs = {
            channel_name: possible_channels[channel_name]
            for channel_name in self.outputs
            if channel_name in possible_channels
        }

        self.age = 0

        self.alive = True
        self.active = True

    def run(self):
        super(Worker, self).run()

        while self.alive:
            try:
                (sources_with_data, [], []) = select.select(
                    [
                        selectable_of(source)
                        for source in self.sources.values()
                    ], [], [],
                    self.config.heartbeat_interval
                )

                while self.active and len(sources_with_data) > 0:
                    source_with_data = sources_with_data.pop(0)

                    for source_name, input_source in self.sources.iteritems():
                        if source_with_data == selectable_of(input_source):
                            if getattr(input_source, "get_nowait", None):
                                try:
                                    payload = input_source.get_nowait()
                                except Queue.Empty:
                                    continue
                                self.handlers[source_name](payload)
                            else:
                                self.handlers[source_name]()

                self.heartbeat()
                self.age += 1

            except select.error as e:
                if e.args[0] not in [errno.EAGAIN, errno.EINTR]:
                    raise
            except OSError as e:
                if e.errno not in [errno.EAGAIN, errno.EINTR]:
                    raise
            except IOError as e:
                if e.errno not in [errno.EAGAIN, errno.EINTR, errno.EBADF]:
                    raise
            except KeyboardInterrupt:
                self.wind_down_gracefully()
            except Exception:
                self.logger.exception("Unhandled error during run loop")
                sys.exit(-1)

    def toggle_active(self, *args):
        self.active = not self.active

    def wind_down_gracefully(self, *args):
        self.alive = False
        self.active = False

    def wind_down_immediately(self, *args):
        self.alive = False
        self.active = False
        sys.exit(0)

    def heartbeat(self):
        if self.pid == os.getpid() and self.parent_pid != os.getppid():
            self.logger.info("Parent process changed! winding down.")
            self.alive = False


def selectable_of(input_source):
    if hasattr(input_source, "_reader"):
        return input_source._reader
    elif hasattr(input_source, "socket"):
        return input_source.socket
    else:
        return input_source



import logging
import os
import signal

from setproctitle import setproctitle


class Proc(object):

    signal_map = {}

    def __init__(self):
        self.logger = logging.getLogger(self.__module__)

        self.pid = None
        self.parent_pid = None

    @property
    def name(self):
        return self.__class__.__name__.lower()

    def setup(self):
        self.pid = os.getpid()
        self.parent_pid = os.getppid()
        self.setup_signals()
        setproctitle("rotterdam: %s" % self.name)

    def run(self):
        self.setup()
        self.logger.info("Starting %s (%d)", self.name, int(self.pid))

    def setup_signals(self):
        for signal_name, handler_name in self.signal_map.iteritems():
            signal.signal(
                getattr(signal, "SIG%s" % signal_name.upper()),
                getattr(self, handler_name)
            )



import time

from .payload import Payload
from .worker import Worker
from .features import is_available


if is_available("concurrency"):
    from gevent import monkey, pool


class Consumer(Worker):

    source_handlers = {
        "ready": "spawn_job"
    }
    outputs = ['results']

    def __init__(self, *args, **kwargs):
        super(Consumer, self).__init__(*args, **kwargs)

        self.pool = None

    def setup(self):
        super(Consumer, self).setup()

        if "concurrency" in self.config:
            monkey.patch_all(thread=False)
            self.pool = pool.Pool(self.config.concurrency)

    def spawn_job(self, payload):
        if self.pool is not None:
            self.pool.spawn(self.run_job, payload)
        else:
            self.run_job(payload)

    def run_job(self, payload):
        self.logger.debug("Job started", extra={"payload": payload})
        start_time = time.time()

        try:
            job = Payload.deserialize(payload)
        except:
            self.logger.exception("Exception when loading job!")
        else:
            try:
                job.run()
            except:
                self.logger.exception("Exception when running job!")

        end_time = time.time()

        self.logger.debug(
            "Job completed in %0.2fs seconds", end_time - start_time
        )
        self.outputs['results'].put(
            {"payload": payload, "time": end_time - start_time}
        )



import datetime
import json

import dateutil.parser
import pytz


TIMESTAMP_PREFIX = "@T-"
INTERVAL_PREFIX = "@I-"


class DateAwareJSONEncoder(json.JSONEncoder):

    def default(self, value):
        if isinstance(value, datetime.datetime):
            return TIMESTAMP_PREFIX + value.isoformat()
        elif isinstance(value, datetime.timedelta):
            return INTERVAL_PREFIX + str(value.total_seconds())

        return super(DateAwareJSONEncoder, self).default(value)


class DateAwareJSONDecoder(json.JSONDecoder):

    def decode(self, value):
        return self.convert(json.loads(value))

    def convert(self, val):
        if isinstance(val, basestring) and val.startswith(TIMESTAMP_PREFIX):
            return dateutil.parser.parse(
                val[len(TIMESTAMP_PREFIX):], tzinfos={"UTC": pytz.utc}
            )
        elif isinstance(val, basestring) and val.startswith(INTERVAL_PREFIX):
            return datetime.timedelta(
                seconds=float(val[len(INTERVAL_PREFIX):])
            )
        elif isinstance(val, dict):
            val = {
                key: self.convert(value)
                for key, value in val.iteritems()
            }
        elif isinstance(val, list):
            val = map(self.convert, val)

        return val



import argparse
import ConfigParser
import importlib
import os
import pkg_resources


version = pkg_resources.get_distribution("rotterdam").version

import settings


DEFAULT_CONFIG_FILENAME = "/etc/rotterdam.conf"


class UnknownConfigError(Exception):
    pass


class Config(object):

    def __init__(self):
        self.settings = {}
        self.parser = None

    @classmethod
    def create(cls, settings_module_name):
        cfg = cls()

        try:
            cfg.load_settings_module(settings_module_name)
        except ImportError:
            raise UnknownConfigError(settings_module_name)

        cfg.create_parser()

        return cfg

    def load_settings_module(self, settings_module_name):
        importlib.import_module("." + settings_module_name, settings.__name__)

        self.load_settings()

    def load_settings(self):
        self.settings = {
            setting_instance.name: setting_instance
            for setting_instance in
            [setting_class() for setting_class in settings.classes]
        }

    def create_parser(self):
        parser = argparse.ArgumentParser()
        parser.add_argument(
            "-v", "--version", action="version",
            version="Rotterdam (version " + version + ")\n",
            help="Print the version and exit"
        )
        parser.add_argument("args", nargs="*", help=argparse.SUPPRESS)

        sorted_settings = sorted(self.settings, key=self.settings.__getitem__)

        for setting_name in sorted_settings:
            self.settings[setting_name].add_option(parser)

        self.parser = parser

    def load(self):
        args = self.parser.parse_args()

        if "config_file" in vars(args) and args.config_file:
            self.load_from_file(args.config_file)
        elif "ROTTERDAM_CONFIG_FILE" in os.environ:
            self.load_from_file(os.environ["ROTTERDAM_CONFIG_FILE"])
        elif os.path.exists(DEFAULT_CONFIG_FILENAME):
            self.load_from_file(DEFAULT_CONFIG_FILENAME)

        for arg_name, arg_value in vars(args).iteritems():
            if arg_value is None or arg_name == "args":
                continue

            try:
                self.set(arg_name.lower(), arg_value)
            except AttributeError:
                pass

        return args

    def load_from_file(self, filename):
        parser = ConfigParser.SafeConfigParser()
        parser.read(filename)

        for arg_name, arg_value in parser.items("rotterdam"):
            if arg_name in self.settings:
                self.set(arg_name, arg_value)

    def set(self, name, value):
        if name not in self.settings:
            raise AttributeError("No configuration setting for: %s" % name)

        self.settings[name].set(value)

    def __getattr__(self, name):
        if name not in self.settings:
            raise AttributeError("No configuration setting for: %s" % name)

        return self.settings[name].get()

    def __setattr__(self, name, value):
        if name != "settings" and name in self.settings:
            raise AttributeError("Setting attributes are read-only.")

        super(Config, self).__setattr__(name, value)

    def __contains__(self, name):
        return name in self.settings and self.settings[name].get() is not None

    def __str__(self):
        return str(self.settings)



import itertools
import os
import time
import types


def get_script_content(command):
    current_path = os.path.dirname(__file__)

    content = ""
    with open("%s/lua/%s.lua" % (current_path, command)) as fd:
        content = fd.read()

    return content


def add_qadd(client):
    content = get_script_content("qadd")

    method = client.register_script(content)

    def qadd(self, queue, when, job_key, job_payload):
        return method(
            keys=[
                "rotterdam:" + queue + ":scheduled",
                "rotterdam:" + queue + ":ready",
                "rotterdam:" + queue + ":working",
                "rotterdam:" + queue + ":jobs:pool"
            ],
            args=[time.time(), when, job_key, job_payload],
            client=self
        )

    client.qadd = types.MethodType(qadd, client)


def add_qpop(client):
    content = get_script_content("qpop")

    method = client.register_script(content)

    def qpop(self, queues, cutoff, maxitems):

        return method(
            keys=list(itertools.chain.from_iterable(
                [
                    "rotterdam:" + queue + ":scheduled",
                    "rotterdam:" + queue + ":working",
                    "rotterdam:" + queue + ":jobs:pool"
                ]
                for queue in queues
            )),
            args=[time.time(), cutoff, maxitems],
            client=self
        )

    client.qpop = types.MethodType(qpop, client)


def add_qfinish(client):
    content = get_script_content("qfinish")

    method = client.register_script(content)

    def qfinish(self, queue, *job_keys):
        args = [time.time()]
        args.extend(job_keys)
        return method(
            keys=[
                "rotterdam:" + queue + ":working",
                "rotterdam:" + queue + ":done",
                "rotterdam:" + queue + ":jobs:pool"
            ],
            args=args,
            client=self
        )

    client.qfinish = types.MethodType(qfinish, client)


def extend_redis(client):
    add_qadd(client)
    add_qpop(client)
    add_qfinish(client)



import Queue
import time

from .worker import Worker
from .payload import Payload


class Arbiter(Worker):

    signal_map = {
        "ttin": "expand_capacity",
        "ttou": "contract_capacity"
    }
    source_handlers = {
        'results': 'handle_finished_job'
    }
    outputs = ['ready']

    def setup(self):
        super(Arbiter, self).setup()

        self.multiplier = 1

        if "concurrency" in self.config:
            self.multiplier = self.config.concurrency

        self.capacity = self.config.num_consumers * self.multiplier

        self.logger.debug(
            "Arbitrating jobs for: %s", ",".join(self.config.queues)
        )

    def heartbeat(self):
        super(Arbiter, self).heartbeat()

        payloads = self.redis.qpop(
            self.config.queues,
            int(time.time()), self.capacity
        )

        for payload in payloads:
            try:
                self.outputs['ready'].put_nowait(payload)
                self.capacity -= 1
                self.logger.debug("Queued job", extra={"payload": payload})
            except Queue.Full:
                break

    def handle_finished_job(self, result):
        self.capacity += 1

        job = Payload.deserialize(result['payload'])
        self.redis.qfinish(job.queue_name, job.unique_key)

    def expand_capacity(self, *args):
        self.capacity += self.multiplier
        self.logger.debug("capacity set to %d", self.capacity)

    def contract_capacity(self, *args):
        self.capacity -= self.multiplier
        self.logger.debug("capacity set to %d", self.capacity)



import datetime
import errno
import json
import socket
import time

from .serialization import DateAwareJSONEncoder
from .exceptions import (
    InvalidPayload, NoSuchJob, ConnectionError, JobEnqueueError
)

SOCKET_BUFFER_SIZE = 4096


class Rotterdam(object):

    def __init__(self, host, port=8765):
        self.host = host
        self.port = port

        self.socket = None

        self.time_offset = None

    @property
    def connected(self):
        return self.socket and self.socket.fileno()

    def connect(self):
        if self.connected:
            return

        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.connect((self.host, self.port))

    def enqueue_in(self, seconds, func, *args, **kwargs):
        self.time_offset = datetime.timedelta(seconds=seconds)
        self.enqueue(func, *args, **kwargs)

    def enqueue_at(self, time_offset, func, *args, **kwargs):
        self.time_offset = time_offset
        self.enqueue(func, *args, **kwargs)

    def enqueue(self, func, *args, **kwargs):
        payload = {
            "args": args,
            "kwargs": kwargs
        }

        if self.time_offset:
            payload['when'] = time.mktime(
                (datetime.datetime.utcnow() + self.time_offset).timetuple()
            )
        self.time_offset = None

        module, func = extract_module_and_func(func)

        payload['module'] = module
        payload['func'] = func

        self.connect()

        self.send_payload(payload)

        response = self.read_response()

        if not response:
            raise JobEnqueueError("empty response")

        response = json.loads(response)

        if response['status'] != "ok":
            if response["message"] == "no such job":
                raise NoSuchJob
            elif response["message"] == "invalid payload":
                raise InvalidPayload
            else:
                raise JobEnqueueError(response["message"])

    def send_payload(self, payload):
        try:
            self.socket.sendall(
                json.dumps(payload, cls=DateAwareJSONEncoder)
                + "\n"
            )
        except IOError, e:
            self.disconnect()
            raise ConnectionError(
                "Error sending job to %s:%s, %s" % (
                    self.host,
                    self.port,
                    e.args[1] if len(e.args) > 1 else e.args[0]
                )
            )

    def read_response(self):
        response = ''
        while True:
            try:
                chunk = self.socket.recv(SOCKET_BUFFER_SIZE)
                if chunk:
                    response = response + chunk
            except socket.error as e:
                if e.errno not in [errno.EAGAIN, errno.EINTR]:
                    raise

            if not response or response.endswith("\n"):
                break

        self.disconnect()

        return response

    def disconnect(self):
        if not self.socket:
            return

        try:
            self.socket.close()
        except socket.error:
            pass
        self.socket = None

    def __del__(self):
        if self.connected:
            self.disconnect()


def extract_module_and_func(func):
    if isinstance(func, basestring):
        module, func = func.split(":")
    else:
        module = func.__module__
        func = func.__name__

    return module, func



class RotterdamError(Exception):
    pass


class NoSuchJob(RotterdamError):
    pass


class ConnectionError(RotterdamError):
    pass


class InvalidPayload(RotterdamError):
    pass


class JobEnqueueError(RotterdamError):
    pass



from setuptools import setup

version_info = (0, 5, 5)

__version__ = ".".join(str(point) for point in version_info)

setup(
    name="rotterdam",
    version=__version__,
    description=(
        "Simple asynchronous job queue via redis."
    ),
    author="William Glass",
    author_email="william.glass@gmail.com",
    url="http://github.com/wglass/rotterdam",
    packages=["rotterdam"],
    include_package_data=True,
    package_data={
        'rotterdam': ['lua/*.lua']
    },
    install_requires=[
        "python-dateutil",
        "pytz"
    ],
    extras_require={
        "server": [
            "setproctitle",
            "redis"
        ]
    },
    tests_require=[
        "mock",
        "nose"
    ],
    entry_points={
        "console_scripts": [
            "rotterdam = rotterdam.scripts.server:run [server]",
            "rotterdamctl = rotterdam.scripts.controller:run [server]",
        ]
    }
)



import errno
import os
import signal
import sys


class WorkerSet(object):

    def __init__(self, master, worker_class):
        self.master = master
        self.worker_class = worker_class
        self.workers = {}

    @property
    def count(self):
        return len(self.workers)

    @count.setter
    def count(self, new_size):
        while len(self.workers) > new_size:
            self.remove_worker()
        while len(self.workers) < new_size:
            self.add_worker()

    def add_worker(self):
        worker = self.worker_class(self.master)

        pid = os.fork()

        if pid != 0:
            self.workers[pid] = worker
            return

        try:
            worker.run()
            sys.exit(0)
        except SystemExit:
            raise
        except:
            self.master.logger.exception(
                "Unhandled exception in %s process", worker.name
            )
            sys.exit(-1)
        finally:
            self.master.logger.info("%s process exiting", worker.name)

    def remove_worker(self):
        (oldest_worker_pid, worker) = sorted(
            self.workers.items(),
            key=lambda i: i[1].age
        ).pop(0)

        self.workers.pop(oldest_worker_pid)
        self.send_signal(signal.SIGQUIT, oldest_worker_pid)

    def broadcast(self, signal):
        for worker_pid in self.workers:
            self.send_signal(signal, worker_pid)

    def send_signal(self, signal, worker_pid):
        try:
            os.kill(worker_pid, signal)
        except OSError as error:
            if error.errno == errno.ESRCH:
                try:
                    self.workers.pop(worker_pid)
                except KeyError:
                    return
            raise

    def regroup(self, regenerate=True):
        exited_worker_pids = []
        for worker_pid in self.workers:
            try:
                pid, status = os.waitpid(worker_pid, os.WNOHANG)
                if pid == worker_pid:
                    exited_worker_pids.append(worker_pid)
            except OSError as e:
                if e.errno == errno.ECHILD:
                    self.workers.pop(worker_pid)
                raise

        for worker_pid in exited_worker_pids:
            try:
                self.workers.pop(worker_pid)
            except KeyError:
                pass
            if regenerate:
                self.add_worker()



import errno
import os
import signal
import sys
import time
from multiprocessing import queues

import redis
import setproctitle

from .connection import Connection
from .proc import Proc
from .arbiter import Arbiter
from .injector import Injector
from .consumer import Consumer
from .worker_set import WorkerSet
from .redis_extensions import extend_redis


class Master(Proc):

    signal_map = {
        "hup": "reload_config",
        "usr1": "relaunch",
        "ttin": "expand_consumers",
        "ttou": "contract_consumers",
        "tstp": "halt_current_jobs",
        "term": "wind_down_gracefully",
        "quit": "wind_down_immediately",
        "chld": "handle_worker_exit"
    }

    def __init__(self, config):
        super(Master, self).__init__()

        self.pid_file_path = None
        self.reexec_pid = 0

        self.config = config

        self.injectors = WorkerSet(self, Injector)
        self.arbiters = WorkerSet(self, Arbiter)
        self.consumers = WorkerSet(self, Consumer)

        self.conn = None
        self.redis = None

        self.ready_queue = None
        self.results_queue = None

        self.wind_down_time = None

        args = sys.argv[:]
        args.insert(0, sys.executable)

        self.launch_context = {
            "executable": sys.executable,
            "args": args,
            "cwd": os.getcwd()
        }

    def load_config(self):
        self.config.load()
        self.pid_file_path = self.config.pid_file

    def setup_pid_file(self):
        if not os.path.isdir(os.path.dirname(self.pid_file_path)):
            raise RuntimeError(
                "%s does not exist! Can't create pid file" % self.pid_file_path
            )

        if os.path.exists(self.pid_file_path):
            raise RuntimeError(
                (
                    "pid file (%s) already exists! \n"
                    + "If no %s process is running it could be stale."
                ) % (self.name, self.pid_file_path)
            )

        fd = open(self.pid_file_path, 'w')
        try:
            fd.write(u"%s\n" % self.pid)
        except Exception:
            os.unlink(self.pid_file_path)
            raise
        else:
            fd.close()

    def setup_connection(self):
        self.conn = Connection(port=self.config.listen_port)
        self.conn.open()
        self.logger.info("Listening on port %s", self.config.listen_port)

    def setup_redis(self):
        if ":" in self.config.redis_host:
            host, port = self.config.redis_host.split(":")
            self.redis = redis.Redis(host=host, port=port)
        else:
            self.redis = redis.Redis(host=self.config.redis_host)

        extend_redis(self.redis)

    def setup_ipc_queues(self):
        self.ready_queue = queues.Queue()
        self.results_queue = queues.Queue()

    def setup(self):
        super(Master, self).setup()
        self.load_config()
        self.setup_pid_file()
        self.setup_ipc_queues()
        self.setup_connection()
        self.setup_redis()

    def run(self):
        super(Master, self).run()
        self.injectors.count = 2
        self.arbiters.count = 1
        self.consumers.count = self.config.num_consumers

        while True:
            try:
                signal.pause()
            except KeyboardInterrupt:
                self.wind_down_gracefully()
            except SystemExit:
                raise
            except Exception:
                self.logger.exception("Error during main loop!")
                self.wind_down_immediately()
                sys.exit(-1)

    def expand_consumers(self, expansion_signal, *args):
        new_count = self.consumers.count + 1
        self.logger.info("Expanding number of consumers to %d", new_count)
        self.consumers.count = new_count
        self.arbiters.broadcast(expansion_signal)

    def contract_consumers(self, contraction_signal, *args):
        if self.consumers.count <= 1:
            self.logger.info(
                "Ignoring contraction, number of consumers already at %d",
                self.consumers.count
            )
            return

        new_count = self.consumers.count - 1
        self.logger.info("Contracting number of consumers to %d", new_count)
        self.consumers.count = new_count
        self.arbiters.broadcast(contraction_signal)

    def reload_config(self, *args):
        self.logger.info("Reloading config")
        old_port = self.config.listen_port

        self.load_config()

        if self.config.listen_port != old_port:
            self.conn.close()
            self.setup_connection()

        self.injectors.count = 2
        self.arbiters.count = 1
        self.consumers.count = self.config.num_consumers

    def relaunch(self, *args):
        os.rename(
            self.pid_file_path,
            self.pid_file_path + ".old." + str(self.pid)
        )

        self.reexec_pid = os.fork()

        if self.reexec_pid != 0:
            self.pid_file_path += ".old." + str(self.pid)
            setproctitle.setproctitle("rotterdam: old %s" % self.name)
            self.wind_down_gracefully()
            return

        os.environ["ROTTERDAM_SOCKET_FD"] = str(self.conn.socket.fileno())

        os.chdir(self.launch_context["cwd"])

        os.execvpe(
            self.launch_context["executable"],
            self.launch_context["args"],
            os.environ
        )

    def broadcast(self, sig):
        self.injectors.broadcast(sig)
        self.arbiters.broadcast(sig)
        self.consumers.broadcast(sig)

    def regroup(self, regenerate=True):
        self.injectors.regroup(regenerate=regenerate)
        self.arbiters.regroup(regenerate=regenerate)
        self.consumers.regroup(regenerate=regenerate)

    def handle_worker_exit(self, *args):
        if not self.wind_down_time:
            self.regroup()
            return

        self.regroup(regenerate=False)
        if (
                self.injectors.count == 0
                and self.arbiters.count == 0
                and self.consumers.count == 0
        ):
            try:
                os.unlink(self.pid_file_path)
            except OSError as e:
                if e.errno == errno.ENOENT:
                    pass
            sys.exit(0)
        else:
            time.sleep(self.wind_down_time - time.time())
            self.broadcast(signal.SIGKILL)

    def halt_current_jobs(self, *args):
        self.logger.info("Toggling job processing")
        self.consumers.broadcast(signal.SIGTSTP)

    def wind_down_gracefully(self, *args):
        self.wind_down(graceful=True)

    def wind_down_immediately(self, *args):
        self.wind_down(graceful=False)

    def wind_down(self, graceful=True):
        if graceful:
            self.logger.info("Winding down gracefully.")
        else:
            self.logger.info("Winding down IMMEDIATELY.")
        self.conn.close()
        self.wind_down_time = (time.time() + self.config.shutdown_grace_period)
        self.broadcast(signal.SIGTERM if graceful else signal.SIGQUIT)



import hashlib
import json
import logging
import os
import random
import time

from .serialization import DateAwareJSONDecoder, DateAwareJSONEncoder
from .exceptions import NoSuchJob, InvalidPayload


logger = logging.getLogger(__name__)


class Payload(object):

    def __init__(self):
        self.module = None
        self.func = None
        self.args = None
        self.kwargs = None
        self.call = None

        self.when = None

        self.unique_key = None

    @classmethod
    def deserialize(cls, message):
        try:
            payload = json.loads(message, cls=DateAwareJSONDecoder)
        except ValueError:
            logger.exception("Error when loading json")
            raise InvalidPayload

        instance = cls.import_func(payload['module'], payload['func'])
        instance.module = payload['module']
        instance.func = payload['func']
        instance.args = payload.get('args', [])
        instance.kwargs = payload.get('kwargs', {})
        instance.unique_key = payload.get("unique_key")
        instance.when = payload.get("when", time.time())

        metadata = instance.call.job_metadata
        instance.queue_name = metadata['queue_name']
        if metadata['delay']:
            instance.when += metadata['delay'].total_seconds()

        if not instance.unique_key:
            instance.determine_unique_key()

        return instance

    @classmethod
    def import_func(cls, module, func):
        instance = cls()
        try:
            module = __import__(module, fromlist=func)
            instance.call = getattr(module, func)
        except (KeyError, ImportError, AttributeError):
            raise NoSuchJob

        return instance

    def determine_unique_key(self):
        uniques = [self.module, self.func, self.queue_name]
        if self.args:
            uniques.extend(self.args)
        if self.kwargs:
            uniques.extend([
                name + "=" + str(value)
                for name, value in self.kwargs.iteritems()
            ])
        if not self.call.job_metadata['unique']:
            uniques += [time.time(), os.getpid(), random.random()]

        uniqueness = hashlib.md5()
        for unique in uniques:
            uniqueness.update(str(unique))

        self.unique_key = uniqueness.hexdigest()

    def serialize(self):
        return json.dumps({
            'when': self.when,
            'unique_key': self.unique_key,
            'module': self.module,
            'func': self.func,
            'args': self.args,
            'kwargs': self.kwargs
        }, cls=DateAwareJSONEncoder)

    def run(self):
        return self.call(*self.args, **self.kwargs)

    def __repr__(self):
        arg_string = ", ".join([str(arg) for arg in self.args])
        kwarg_string = ", ".join([
            "%s=%s" % (name, val)
            for name, val in self.kwargs.iteritems()
        ])
        if arg_string and kwarg_string:
            kwarg_string = ", " + kwarg_string
        return "%(module)s.%(func)s(%(args)s%(kwargs)s)" % {
            "module": self.module,
            "func": self.func,
            "args": arg_string,
            "kwargs": kwarg_string
        }



import functools


def job(queue_name, unique=False, delay=None):

    def inner(fn):
        fn.job_metadata = {
            "queue_name": queue_name,
            "unique": unique,
            "delay": delay
        }

        @functools.wraps(fn)
        def wrapper(*wrapped_args, **wrapped_kwargs):
            return fn(*wrapped_args, **wrapped_kwargs)

        return wrapper

    return inner



from .worker import Worker


class Injector(Worker):

    source_handlers = {
        "connection": "handle_incoming_jobs"
    }

    def handle_incoming_jobs(self):
        for job in self.sources['connection']:
            self.logger.debug("job recieved: %s", job)

            self.redis.qadd(
                job.queue_name,
                job.when,
                job.unique_key,
                job.serialize()
            )
